"""Makes separate collision meshes for each link in the URDF."""

import argparse
import logging
import shutil
from pathlib import Path

from onshape.passes.utils import iter_meshes
from onshape.utils.mesh import COLLISION_SUFFIX

logger = logging.getLogger(__name__)


def separate_collision_mesh(visual_mesh_path: Path) -> Path:
    collision_mesh_name = f"{visual_mesh_path.stem}{COLLISION_SUFFIX}{visual_mesh_path.suffix}"
    collision_mesh_path = visual_mesh_path.parent / collision_mesh_name
    if collision_mesh_path.exists():
        logger.warning("Collision mesh %s already exists for %s", collision_mesh_path, visual_mesh_path)
    shutil.copy2(visual_mesh_path, collision_mesh_path)
    return collision_mesh_path


def separate_collision_meshes_in_urdf(urdf_path: Path) -> None:
    """Creates separate collision mesh files for each link in a URDF.

    For each link that has both visual and collision meshes pointing to the same file,
    creates a separate collision mesh file and updates the URDF accordingly.

    Args:
        urdf_path: The path to the URDF file.
    """
    num_separated = 0
    for _, (_, visual_mesh_path), (collision_link_elem, collision_mesh_path) in iter_meshes(
        urdf_path,
        save_when_done=True,
    ):
        if visual_mesh_path is None or collision_link_elem is None or collision_mesh_path is None:
            continue

        if visual_mesh_path == collision_mesh_path:
            new_collision_path = separate_collision_mesh(visual_mesh_path)
            rel_filename = new_collision_path.relative_to(urdf_path.parent).as_posix()
            collision_link_elem.attrib["filename"] = rel_filename
            num_separated += 1

    if num_separated > 0:
        logger.info(
            "Created %d separate collision meshes for %s",
            num_separated,
            urdf_path.name,
        )


def main() -> None:
    parser = argparse.ArgumentParser(description="Create separate collision meshes for each link in a URDF.")
    parser.add_argument("urdf_path", type=Path, help="The path to the URDF file.")
    args = parser.parse_args()

    separate_collision_meshes_in_urdf(args.urdf_path)


if __name__ == "__main__":
    # python -m onshape.passes.separate_collision_meshes
    main()



# ruff: noqa: N815
"""Defines the schema for parts APIs."""

import functools
from typing import Any

import numpy as np
from pydantic import BaseModel


class PartProperty(BaseModel):
    name: str
    value: Any


class PartMetadata(BaseModel):
    jsonType: str
    isMesh: bool
    partId: str
    isFlattenedBody: bool
    partType: str
    meshState: int
    properties: list[PartProperty]
    href: str

    @functools.cached_property
    def property_map(self) -> dict[str, Any]:
        return {prop.name: prop.value for prop in self.properties}


class PrincipleAxis(BaseModel):
    x: float
    y: float
    z: float

    @property
    def array(self) -> np.ndarray:
        return np.array([self.x, self.y, self.z])


class PartBody(BaseModel):
    mass: list[float]
    volume: list[float]
    periphery: list[float]
    centroid: list[float]
    inertia: list[float]
    hasMass: bool
    massMissingCount: int
    principalInertia: list[float]
    principalAxes: list[PrincipleAxis]

    @property
    def principal_inertia(self) -> np.ndarray:
        return np.array(self.principalInertia)

    @property
    def principal_axes(self) -> np.matrix:
        return np.matrix(np.array([axis.array for axis in self.principalAxes]))

    def principal_axes_in_frame(self, frame: np.ndarray) -> np.ndarray:
        return frame[:3, :3] @ self.principal_axes

    @property
    def inertia_matrix(self) -> np.matrix:
        return np.matrix(np.array(self.inertia[:9]).reshape(3, 3))

    def inertia_in_frame(self, frame: np.matrix) -> np.matrix:
        frame_r = frame[:3, :3].T
        return frame_r * self.inertia_matrix * frame_r.T

    @property
    def center_of_mass(self) -> tuple[float, float, float]:
        return (self.centroid[0], self.centroid[1], self.centroid[2])

    def center_of_mass_in_frame(self, frame: np.ndarray) -> tuple[float, float, float]:
        com = np.matrix(list(self.center_of_mass) + [1.0])
        com_in_frame = (frame * com.T)[:3]
        return (float(com_in_frame[0, 0]), float(com_in_frame[1, 0]), float(com_in_frame[2, 0]))


class PartDynamics(BaseModel):
    microversionId: str
    bodies: dict[str, PartBody]


class ThumbnailSize(BaseModel):
    size: str
    href: str
    mediaType: str
    uniqueId: str | None
    viewOrientation: str
    renderMode: str
    sheetName: str | None


class ThumbnailInfo(BaseModel):
    sizes: list[ThumbnailSize]
    secondarySizes: list[ThumbnailSize] | None
    id: str | None
    href: str



"""Tests geometry functions."""

from onshape.utils.geometry import get_mesh_convex_hull, scale_mesh
from onshape.utils.mesh import load_file


def test_scale_mesh() -> None:
    mesh = load_file("tests/data/random.stl")

    # Tests getting the mesh convex hull. The particular mesh is convex, so
    # the convex hull should be the same as the original mesh.
    hull_mesh = get_mesh_convex_hull(mesh)
    assert len(hull_mesh.faces) <= len(mesh.faces)

    # Tests scaling the mesh.
    scaled_mesh = scale_mesh(mesh, 2.0)
    assert not (scaled_mesh.points == mesh.points).all()
    assert (scaled_mesh.faces == mesh.faces).all()

    # Tests scaling the mesh about the origin.
    scaled_mesh = scale_mesh(mesh, 2.0, about_origin=True)
    assert not (scaled_mesh.points == mesh.points).all()
    assert (scaled_mesh.faces == mesh.faces).all()



"""Performs some simple checks for writing URDF files."""

from pathlib import Path

from onshape.formats import urdf


def test_urdf(tmpdir: Path) -> None:
    links: list[urdf.Link | urdf.BaseJoint] = [
        urdf.Link(
            "first",
            urdf.VisualLink(
                origin=urdf.Origin((0, 0, 0), (0, 0, 0)),
                geometry=urdf.BoxGeometry((1.5, 1.0, 1.0)),
                material=urdf.Material.from_color("red"),
            ),
            urdf.CollisionLink(
                origin=urdf.Origin((0, 0, 0), (0, 0, 0)),
                geometry=urdf.BoxGeometry((1.0, 1.0, 1.0)),
            ),
            urdf.InertialLink(
                mass=1.0,
                inertia=urdf.Inertia(1.0, 1.0, 1.0, 1.0, 1.0, 1.0),
                origin=urdf.Origin((0, 0, 0), (0, 0, 0)),
            ),
        ),
        urdf.Link(
            "second",
            urdf.VisualLink(
                origin=urdf.Origin((0, 0, 0), (0, 0, 0)),
                geometry=urdf.MeshGeometry("mesh.stl"),
                material=urdf.Material.from_color("blue"),
            ),
            urdf.CollisionLink(
                origin=urdf.Origin((0, 0, 0), (0, 0, 0)),
                geometry=urdf.SphereGeometry(1.0),
            ),
            urdf.InertialLink(
                mass=1.0,
                inertia=urdf.Inertia(1.0, 1.0, 1.0, 1.0, 1.0, 1.0),
                origin=urdf.Origin((0, 0, 0), (0, 0, 0)),
            ),
        ),
        urdf.RevoluteJoint(
            name="joint",
            parent="first",
            child="second",
            origin=urdf.Origin((0, 0, 0), (0, 0, 0)),
            limits=urdf.JointLimits(1.0, 1.0, -1.0, 1.0),
            axis=urdf.Axis((0, 0, 1)),
        ),
    ]

    robot = urdf.Robot("robot", links)
    save_path = Path(tmpdir / "robot.urdf")
    robot.save(save_path)

    # Check that the saved URDF is a valid XML file.
    assert save_path.exists()
    assert save_path.stat().st_size > 0
    assert save_path.suffix == ".urdf"
    assert save_path.read_text()



"""Simple script to interact with a URDF in Mujoco."""

import argparse
import logging
from pathlib import Path
from typing import Sequence

from onshape.utils.logging import configure_logging

logger = logging.getLogger(__name__)


def mujoco_main(args: Sequence[str] | None = None) -> None:
    configure_logging()

    parser = argparse.ArgumentParser(description="Show a MJCF in Mujoco")
    parser.add_argument("mjcf", nargs="?", help="Path to the MJCF file")
    parser.add_argument("--dt", type=float, default=0.01, help="Time step")
    parser.add_argument("--start-height", type=float, default=1.0, help="Initial height of the robot")
    parser.add_argument("--fixed-base", action="store_true", help="Fix the base linkage in place")
    parser.add_argument("--cycle-duration", type=float, default=10, help="Duration of the joint cycling")
    parsed_args = parser.parse_args(args)

    try:
        from mujoco.viewer import launch_from_path
    except ImportError:
        raise ImportError(
            "mujoco and mujoco_viewer are required to run this script; install them "
            "with `pip install mujoco mujoco-python-viewer`"
        )

    # Load the MJCF file
    mjcf_path = Path("robot" if parsed_args.mjcf is None else parsed_args.mjcf)
    if mjcf_path.is_dir():
        try:
            mjcf_path = next(mjcf_path.glob("*.mjcf"))
        except StopIteration:
            raise FileNotFoundError(f"No MJCF files found in {mjcf_path}")
    elif not mjcf_path.is_file():
        raise FileNotFoundError(mjcf_path)

    # Set up the viewer
    launch_from_path(str(mjcf_path))


if __name__ == "__main__":
    # python -m onshape.scripts.mujoco
    mujoco_main()



"""Runs an end-to-end test of the URDF exporter on the Stompy model."""

import asyncio
from pathlib import Path

import pytest

from onshape.onshape.config import ConverterConfig
from onshape.onshape.download import download
from onshape.onshape.postprocess import postprocess
from onshape.utils.logging import configure_logging

# STOMPY_ONSHAPE_URL = (
#     "https://cad.onshape.com/documents/71f793a23ab7562fb9dec82d/w/"
#     "e879bcf272425973f9b3d8ad/e/1a95e260677a2d2d5a3b1eb3"
# )

ONSHAPE_URL = (
    "https://cad.onshape.com/documents/4b3eeb430e3d28511ab9cba8/w/"
    "dc58c6e8cf7c1ee8bd864ee4/e/de14dfcca89a312f32f77d02"
)


@pytest.mark.skip(reason="This test is slow and requires an internet connection")
async def test_e2e(tmpdir: Path) -> None:
    """Runs an end-to-end test of the URDF exporter on the Stompy model.

    Args:
        tmpdir: The temporary directory to save the URDF file.
        ext: The mesh file format.
    """
    config = ConverterConfig(
        document_url=ONSHAPE_URL,
        output_dir=str(tmpdir),
        default_prismatic_joint_effort=100,
        default_prismatic_joint_velocity=5,
        default_revolute_joint_effort=100,
        default_revolute_joint_velocity=5,
        suffix_to_joint_effort={
            "dof_x4_h": 1.5,
            "dof_x4": 1.5,
            "dof_x6": 3,
            "dof_x8": 6,
            "dof_x10": 12,
            "knee_revolute": 13.9,
            "ankle_revolute": 6,
        },
        suffix_to_joint_velocity={},
        disable_mimics=True,
    )

    document_info = await download(
        document_url=config.document_url,
        output_dir=config.output_dir,
        config=config,
    )

    await postprocess(
        urdf_path=document_info.urdf_info.urdf_path,
        config=config,
    )


if __name__ == "__main__":
    # python -m tests.test_e2e
    configure_logging()
    asyncio.run(test_e2e(Path.cwd() / "test_e2e"))



"""Defines functions for post-processing the downloaded URDF."""

import asyncio
import logging
import sys
import tarfile
from dataclasses import dataclass
from pathlib import Path
from typing import Sequence

from onshape.onshape.config import ConverterConfig, PostprocessConfig
from onshape.onshape.download import download
from onshape.passes.add_base_linkage import add_base_linkage
from onshape.passes.add_joint_separation import add_joint_separation
from onshape.passes.add_mjcf import convert_urdf_to_mjcf
from onshape.passes.convert_floats_to_consistent_types import convert_floats_to_consistent_types
from onshape.passes.exclude_collision_meshes import exclude_collision_meshes
from onshape.passes.fix_inertias import fix_inertias
from onshape.passes.flip_joints import flip_joints
from onshape.passes.make_convex_collision_mesh import get_convex_collision_meshes
from onshape.passes.merge_fixed_joints import get_merged_urdf
from onshape.passes.move_collision_meshes import move_collision_meshes
from onshape.passes.remove_extra_meshes import remove_extra_meshes
from onshape.passes.remove_internal_geometries import remove_internal_geometries_from_urdf
from onshape.passes.rotate_joints import rotate_joints
from onshape.passes.separate_collision_meshes import separate_collision_meshes_in_urdf
from onshape.passes.shrink_collision_meshes import shrink_collision_meshes
from onshape.passes.simplify_meshes import get_simplified_urdf
from onshape.passes.sort_sections import sort_sections
from onshape.passes.update_names import update_urdf_names
from onshape.passes.use_collision_meshes_as_visual_meshes import use_collision_meshes_as_visual_meshes
from onshape.passes.utils import iter_meshes
from onshape.utils.logging import configure_logging


@dataclass
class PostprocessedDocument:
    urdf_path: Path
    tar_path: Path


async def postprocess(
    urdf_path: str | Path,
    *,
    config: PostprocessConfig | None = None,
) -> PostprocessedDocument:
    """Combines multiple post-processing steps into a single function.

    Args:
        urdf_path: The path to the downloaded URDF to post-process.
        config: The post-processing configuration.

    Returns:
        The post-processed document.
    """
    urdf_path = Path(urdf_path)
    if config is None:
        config = PostprocessConfig()

    # Updates the names in the URDF.
    if config.update_names:
        update_urdf_names(urdf_path, joint_name_map=config.joint_name_map, link_name_map=config.link_name_map)

    # Merges all fixed joints in the URDF.
    if config.merge_fixed_joints:
        get_merged_urdf(urdf_path, ignore_merging_fixed_joints=config.ignore_merging_fixed_joints)

    # Applies rotations to some joints.
    if config.rotate_joints is not None:
        rotate_joints(urdf_path, config.rotate_joints)

    # Creates separate collision meshes for each link in the URDF.
    if config.separate_collision_meshes:
        separate_collision_meshes_in_urdf(urdf_path)

    # Shrinks some of the collision meshes.
    if config.shrink_collision_meshes is not None:
        shrink_collision_meshes(urdf_path, config.shrink_collision_meshes)

    # Moves some of the collision meshes.
    if config.move_collision_meshes is not None:
        move_collision_meshes(urdf_path, config.move_collision_meshes)

    # Creates separate convex hulls for collision geomtries.
    if config.convex_collision_meshes:
        get_convex_collision_meshes(
            urdf_path,
            max_triangles=config.max_convex_collision_mesh_triangles,
        )

    # Simplifies the meshes in the URDF.
    if config.simplify_meshes:
        get_simplified_urdf(
            urdf_path,
            voxel_size=config.voxel_size,
            max_triangles=config.max_mesh_triangles,
        )

    # Add a small separation between adjacent joints.
    if config.joint_separation_distance is not None:
        add_joint_separation(urdf_path, config.joint_separation_distance)

    # Flips the orientation of specific joints.
    if config.flip_joints is not None:
        flip_joints(urdf_path, config.flip_joints)

    # Fixes the inertias in the URDF.
    if config.fix_inertias:
        fix_inertias(urdf_path, epsilon=config.min_inertia_eigval)

    # Converts floating point numbers to consistent types.
    if config.convert_floats_to_consistent_types:
        convert_floats_to_consistent_types(urdf_path)

    # Sorts the sections in the URDF, putting the joints at the end.
    if config.sort_sections:
        sort_sections(urdf_path)

    # Adds a base linkage.
    if config.add_base_linkage:
        add_base_linkage(urdf_path, base_xyz=config.base_xyz, base_rpy=config.base_rpy)

    # Remove internal triangles from the meshes in the URDF.
    if config.remove_internal_geometries:
        remove_internal_geometries_from_urdf(urdf_path)

    # Excludes collision meshes from the URDF.
    if config.exclude_collision_meshes:
        exclude_collision_meshes(urdf_path, config.exclude_collision_meshes)

    # Use collision meshes as visual meshes.
    if config.use_collision_meshes_as_visual_meshes:
        use_collision_meshes_as_visual_meshes(urdf_path)

    # Adds the MJCF XML to the package.
    paths = [urdf_path]
    if config.add_mjcf:
        for metadata in config.mjcf_metadata or []:
            if metadata.suffix is not None:
                paths.extend(
                    convert_urdf_to_mjcf(
                        urdf_file=urdf_path,
                        mjcf_file=urdf_path.with_suffix(f".{metadata.suffix}.mjcf"),
                        metadata=metadata,
                        joint_metadata=config.joint_metadata,
                        actuator_metadata=config.actuators_metadata,
                    )
                )
            else:
                paths.extend(
                    convert_urdf_to_mjcf(
                        urdf_file=urdf_path,
                        metadata=metadata,
                        joint_metadata=config.joint_metadata,
                        actuator_metadata=config.actuators_metadata,
                    )
                )

    if config.remove_extra_meshes:
        remove_extra_meshes(urdf_path)

    # Combines everything to a single TAR file.
    for _, (_, visual_mesh_path), (_, collision_mesh_path) in iter_meshes(urdf_path):
        for path in list({visual_mesh_path, collision_mesh_path}):
            if path is not None:
                paths.append(path)

    tar_path = urdf_path.with_suffix(".tgz")
    with tarfile.open(tar_path, "w:gz") as tar:
        for path in paths:
            arcname = path.relative_to(urdf_path.parent).as_posix()
            tar.add(path, arcname=arcname)

    return PostprocessedDocument(
        urdf_path=urdf_path,
        tar_path=tar_path,
    )


async def download_and_postprocess_main(args: Sequence[str] | None = None) -> PostprocessedDocument:
    if args is None:
        args = sys.argv[1:]
    config = ConverterConfig.from_cli_args(args)
    configure_logging(level=logging.DEBUG if config.debug else logging.INFO)
    document_info = await download(
        document_url=config.document_url,
        output_dir=config.output_dir,
        config=config,
    )
    config.urdf_path = document_info.urdf_info.urdf_path.as_posix()
    return await postprocess(
        urdf_path=document_info.urdf_info.urdf_path,
        config=config,
    )


async def postprocess_main(args: Sequence[str] | None = None) -> PostprocessedDocument:
    if args is None:
        args = sys.argv[1:]
    config = PostprocessConfig.from_cli_args(args)
    configure_logging(level=logging.DEBUG if config.debug else logging.INFO)
    return await postprocess(
        urdf_path=config.urdf_path,
        config=config,
    )


def sync_main(args: Sequence[str] | None = None) -> None:
    asyncio.run(postprocess_main(args))


if __name__ == "__main__":
    # python -m onshape.onshape.postprocess
    sync_main()



"""Defines utility functions for enabling logging."""

import datetime
import logging
import re
import sys
from typing import Literal

# Show as a transient message.
LOG_PING: int = logging.INFO + 2

# Show as a persistent status message.
LOG_STATUS: int = logging.INFO + 3

RESET_SEQ = "\033[0m"
REG_COLOR_SEQ = "\033[%dm"
BOLD_COLOR_SEQ = "\033[1;%dm"
BOLD_SEQ = "\033[1m"

Color = Literal[
    "black",
    "red",
    "green",
    "yellow",
    "blue",
    "magenta",
    "cyan",
    "white",
    "grey",
    "light-red",
    "light-green",
    "light-yellow",
    "light-blue",
    "light-magenta",
    "light-cyan",
]

COLOR_INDEX: dict[Color, int] = {
    "black": 30,
    "red": 31,
    "green": 32,
    "yellow": 33,
    "blue": 34,
    "magenta": 35,
    "cyan": 36,
    "white": 37,
    "grey": 90,
    "light-red": 91,
    "light-green": 92,
    "light-yellow": 93,
    "light-blue": 94,
    "light-magenta": 95,
    "light-cyan": 96,
}


def color_parts(color: Color, bold: bool = False) -> tuple[str, str]:
    if bold:
        return BOLD_COLOR_SEQ % COLOR_INDEX[color], RESET_SEQ
    return REG_COLOR_SEQ % COLOR_INDEX[color], RESET_SEQ


def uncolored(s: str) -> str:
    return re.sub(r"\033\[[\d;]+m", "", s)


def colored(s: str, color: Color | None = None, bold: bool = False) -> str:
    if color is None:
        return s
    start, end = color_parts(color, bold=bold)
    return start + s + end


def wrapped(
    s: str,
    length: int | None = None,
    space: str = " ",
    spaces: str | re.Pattern = r" ",
    newlines: str | re.Pattern = r"[\n\r]",
    too_long_suffix: str = "...",
) -> list[str]:
    strings = []
    lines = re.split(newlines, s.strip(), flags=re.MULTILINE | re.UNICODE)
    for line in lines:
        cur_string = []
        cur_length = 0
        for part in re.split(spaces, line.strip(), flags=re.MULTILINE | re.UNICODE):
            if length is None:
                cur_string.append(part)
                cur_length += len(space) + len(part)
            else:
                if len(part) > length:
                    part = part[: length - len(too_long_suffix)] + too_long_suffix
                if cur_length + len(part) > length:
                    strings.append(space.join(cur_string))
                    cur_string = [part]
                    cur_length = len(part)
                else:
                    cur_string.append(part)
                    cur_length += len(space) + len(part)
        if cur_length > 0:
            strings.append(space.join(cur_string))
    return strings


def outlined(
    s: str,
    inner: Color | None = None,
    side: Color | None = None,
    bold: bool = False,
    max_length: int | None = None,
    space: str = " ",
    spaces: str | re.Pattern = r" ",
    newlines: str | re.Pattern = r"[\n\r]",
) -> str:
    strs = wrapped(uncolored(s), max_length, space, spaces, newlines)
    max_len = max(len(s) for s in strs)
    strs = [f"{s}{' ' * (max_len - len(s))}" for s in strs]
    strs = [colored(s, inner, bold=bold) for s in strs]
    strs_with_sides = [f"{colored('│', side)} {s} {colored('│', side)}" for s in strs]
    top = colored("┌─" + "─" * max_len + "─┐", side)
    bottom = colored("└─" + "─" * max_len + "─┘", side)
    return "\n".join([top] + strs_with_sides + [bottom])


def show_info(s: str, important: bool = False) -> None:
    if important:
        s = outlined(s, inner="light-cyan", side="cyan", bold=True)
    else:
        s = colored(s, "light-cyan", bold=False)
    sys.stdout.write(s)
    sys.stdout.write("\n")
    sys.stdout.flush()


def show_error(s: str, important: bool = False) -> None:
    if important:
        s = outlined(s, inner="light-red", side="red", bold=True)
    else:
        s = colored(s, "light-red", bold=False)
    sys.stdout.write(s)
    sys.stdout.write("\n")
    sys.stdout.flush()


def show_warning(s: str, important: bool = False) -> None:
    if important:
        s = outlined(s, inner="light-yellow", side="yellow", bold=True)
    else:
        s = colored(s, "light-yellow", bold=False)
    sys.stdout.write(s)
    sys.stdout.write("\n")
    sys.stdout.flush()


def format_timedelta(timedelta: datetime.timedelta, short: bool = False) -> str:
    """Formats a delta time to human-readable format.

    Args:
        timedelta: The delta to format
        short: If set, uses a shorter format

    Returns:
        The human-readable time delta
    """
    parts = []
    if timedelta.days > 0:
        if short:
            parts += [f"{timedelta.days}d"]
        else:
            parts += [f"{timedelta.days} day" if timedelta.days == 1 else f"{timedelta.days} days"]

    seconds = timedelta.seconds

    if seconds > 60 * 60:
        hours, seconds = seconds // (60 * 60), seconds % (60 * 60)
        if short:
            parts += [f"{hours}h"]
        else:
            parts += [f"{hours} hour" if hours == 1 else f"{hours} hours"]

    if seconds > 60:
        minutes, seconds = seconds // 60, seconds % 60
        if short:
            parts += [f"{minutes}m"]
        else:
            parts += [f"{minutes} minute" if minutes == 1 else f"{minutes} minutes"]

    if short:
        parts += [f"{seconds}s"]
    else:
        parts += [f"{seconds} second" if seconds == 1 else f"{seconds} seconds"]

    return ", ".join(parts)


def format_datetime(dt: datetime.datetime) -> str:
    """Formats a datetime to human-readable format.

    Args:
        dt: The datetime to format

    Returns:
        The human-readable datetime
    """
    return dt.strftime("%Y-%m-%d %H:%M:%S")


def camelcase_to_snakecase(s: str) -> str:
    return re.sub(r"([a-z0-9])([A-Z])", r"\1_\2", s).lower()


def snakecase_to_camelcase(s: str) -> str:
    return "".join(word.title() for word in s.split("_"))


def highlight_exception_message(s: str) -> str:
    s = re.sub(r"(\w+Error)", r"\033[1;31m\1\033[0m", s)
    s = re.sub(r"(\w+Exception)", r"\033[1;31m\1\033[0m", s)
    s = re.sub(r"(\w+Warning)", r"\033[1;33m\1\033[0m", s)
    s = re.sub(r"\^+", r"\033[1;35m\g<0>\033[0m", s)
    s = re.sub(r"File \"(.+?)\"", r'File "\033[36m\1\033[0m"', s)
    return s


def is_interactive_session() -> bool:
    return hasattr(sys, "ps1") or hasattr(sys, "ps2") or sys.stdout.isatty() or sys.stderr.isatty()


class ColoredFormatter(logging.Formatter):
    """Defines a custom formatter for displaying logs."""

    RESET_SEQ = "\033[0m"
    COLOR_SEQ = "\033[1;%dm"
    BOLD_SEQ = "\033[1m"

    COLORS: dict[str, Color] = {
        "WARNING": "yellow",
        "INFO": "cyan",
        "DEBUG": "grey",
        "CRITICAL": "yellow",
        "FATAL": "red",
        "ERROR": "red",
        "STATUS": "green",
        "PING": "magenta",
    }

    def __init__(
        self,
        *,
        prefix: str | None = None,
        use_color: bool = True,
    ) -> None:
        asc_start, asc_end = color_parts("grey")
        name_start, name_end = color_parts("blue", bold=True)

        message_pre = [
            "{levelname:^19s}",
            asc_start,
            "{asctime}",
            asc_end,
            " [",
            name_start,
            "{name}",
            name_end,
            "]",
        ]
        message_post = [" {message}"]

        if prefix is not None:
            message_pre += [" ", colored(prefix, "magenta", bold=True)]

        message = "".join(message_pre + message_post)

        super().__init__(message, style="{", datefmt="%Y-%m-%d %H:%M:%S")

        self.use_color = use_color

    def format(self, record: logging.LogRecord) -> str:
        levelname = record.levelname
        if record.levelname and self.use_color and levelname in self.COLORS:
            record.levelname = colored(record.levelname, self.COLORS[levelname], bold=True)
        return logging.Formatter.format(self, record)


def configure_logging(prefix: str | None = None, level: int = logging.INFO) -> None:
    """Instantiates logging.

    This captures logs and reroutes them to the Toasts module, which is
    pretty similar to Python logging except that the API is a lot easier to
    interact with.

    Args:
        prefix: An optional prefix to add to the logger
        level: The logging level
    """
    root_logger = logging.getLogger()

    # Adds new level names.
    logging.addLevelName(LOG_PING, "PING")
    logging.addLevelName(LOG_STATUS, "STATUS")

    # Captures warnings from the warnings module.
    logging.captureWarnings(True)

    # Clears all existing handlers.
    for handler in root_logger.handlers[:]:
        root_logger.removeHandler(handler)

    # Adds new handler.
    stream_handler = logging.StreamHandler(sys.stdout)
    stream_handler.setFormatter(ColoredFormatter(prefix=prefix))
    root_logger.addHandler(stream_handler)

    root_logger.setLevel(level)



"""Defines a utility class for resolving expression values."""

import logging
import math
import re

logger = logging.getLogger(__name__)


class ExpressionResolver:
    def __init__(self, configuration: str) -> None:
        super().__init__()

        # Parses the configuration string to a dictionary.
        self.configuration_parameters: dict[str, str] = {}
        if configuration != "default":
            for kv in configuration.split(";"):
                key, value = kv.split("=")
                self.configuration_parameters[key] = value

    def read_expression(self, expression: str) -> float:
        """Reads an expression and returns a float value.

        Args:
            expression: The expression to read.

        Returns:
            The float value of the expression.
        """
        if expression[0] == "#":
            expression = self.configuration_parameters[expression[1:]]
        if expression[0:2] == "-#":
            expression = "-" + self.configuration_parameters[expression[2:]]

        # Splitting the expression into value and unit.
        parts = re.split(r"[ +*]", expression)
        if len(parts) != 2:
            raise ValueError(f"Invalid expression: {expression}")
        value, unit = parts

        # Checking the unit, returning values in radians and meters.
        match unit:
            case "deg":
                return math.radians(float(value))
            case "radian" | "rad":
                try:
                    val = float(value)
                except ValueError:
                    if value == "(PI)":
                        val = math.pi
                    else:
                        raise ValueError(f"{value} variable isn't supported")
                return float(val)
            case "mm":
                return float(value) / 1000.0
            case "m":
                return float(value)
            case "cm":
                return float(value) / 100.0
            case "in":
                return float(value) * 0.0254
            case "ft":
                return float(value) * 0.3048
            case _:
                raise ValueError(f"{unit} unit isn't supported")



"""Moves collision meshes by specified offsets."""

import logging
import shutil
import xml.etree.ElementTree as ET
from pathlib import Path
from typing import Mapping, Sequence

import numpy as np
import trimesh

from onshape.passes.utils import iter_meshes
from onshape.utils.mesh import COLLISION_SUFFIX

logger = logging.getLogger(__name__)


def separate_collision_mesh(visual_mesh_path: Path) -> Path:
    """Creates a separate collision mesh file from a visual mesh.

    Args:
        visual_mesh_path: Path to the visual mesh file.

    Returns:
        Path to the new collision mesh file.
    """
    collision_mesh_name = f"{visual_mesh_path.stem}{COLLISION_SUFFIX}{visual_mesh_path.suffix}"
    collision_mesh_path = visual_mesh_path.parent / collision_mesh_name
    if collision_mesh_path.exists():
        logger.warning("Collision mesh %s already exists for %s", collision_mesh_path, visual_mesh_path)
    shutil.copy2(visual_mesh_path, collision_mesh_path)
    return collision_mesh_path


def move_collision_mesh(collision_mesh_path: Path, translation: Sequence[float]) -> None:
    """Moves a collision mesh by a specified translation.

    Args:
        collision_mesh_path: Path to the collision mesh file.
        translation: [x, y, z] translation to apply to the mesh.
    """
    mesh = trimesh.load(collision_mesh_path)
    if not isinstance(mesh, trimesh.Trimesh):
        raise ValueError(f"Loaded mesh {collision_mesh_path} is not a Trimesh")
    translation = np.array(translation, dtype=np.float64)
    mesh.apply_translation(translation)
    mesh.export(collision_mesh_path)


def move_collision_meshes(
    urdf_path: Path,
    translations: Mapping[str, Sequence[float]],
) -> None:
    """Moves collision meshes in the URDF by specified translations.

    Args:
        urdf_path: The path to the URDF file.
        translations: A dictionary mapping link names to [x, y, z] translations.
    """
    # Make a copy that we can use to keep track of which links have been processed
    translations = {k: v for k, v in translations.items()}

    num_separated = 0
    for link, (_, visual_mesh_path), (col_link, col_mesh_path) in iter_meshes(
        urdf_path,
        save_when_done=True,
    ):
        name = link.attrib["name"]

        if name not in translations:
            continue

        if col_link is None or col_mesh_path is None:
            raise ValueError(f"No collision mesh found for {name}")

        # If the collision mesh is the same as the visual mesh, we need to make
        # a new collision mesh
        if visual_mesh_path is not None and visual_mesh_path == col_mesh_path:
            col_mesh_path = separate_collision_mesh(visual_mesh_path)
            rel_filename = col_mesh_path.relative_to(urdf_path.parent).as_posix()
            col_link.attrib["filename"] = rel_filename
            num_separated += 1

        # After doing this, we can move the collision mesh geometry
        translation = translations.pop(name)
        if len(translation) != 3:
            raise ValueError(f"Expected translation to be a list of 3 floats, got {translation}")
        logger.info("Moving collision mesh %s by %s", col_mesh_path, translation)
        move_collision_mesh(col_mesh_path, translation)

    if len(translations) > 0:
        missing_keys = list(translations.keys())
        urdf_tree = ET.parse(urdf_path)
        options = [link.attrib["name"] for link in urdf_tree.findall(".//link")]
        raise ValueError(f"The following links were not processed: {missing_keys}. Options are {options}")

    if num_separated > 0:
        logger.info(
            "Created %d separate collision meshes for %s",
            num_separated,
            urdf_path.name,
        )



"""Converts floating point numbers to consistent types."""

import logging
import xml.etree.ElementTree as ET
from pathlib import Path

from onshape.formats.common import save_xml

logger = logging.getLogger(__name__)


def flstr(value: float) -> str:
    if abs(value) < 1e-6:
        return "0"
    if abs(value - 1) < 1e-6:
        return "1"
    return f"{value:.6f}"


def flstrs(values: str) -> str:
    return " ".join(flstr(float(value)) for value in values.split())


def convert_floats_to_consistent_types(urdf_path: Path) -> None:
    """Converts floating point numbers to consistent types."""
    tree = ET.parse(urdf_path)
    root = tree.getroot()

    # Fix up joint.
    for joint in root.findall("joint"):
        logger.info("Converting joint %s", joint.get("name"))
        if (origin := joint.find("origin")) is not None:
            if (xyz := origin.get("xyz")) is not None:
                origin.set("xyz", flstrs(xyz))
            if (rpy := origin.get("rpy")) is not None:
                origin.set("rpy", flstrs(rpy))
        if (limit := joint.find("limit")) is not None:
            if (lower := limit.get("lower")) is not None:
                limit.set("lower", flstr(float(lower)))
            if (upper := limit.get("upper")) is not None:
                limit.set("upper", flstr(float(upper)))

    # Fix up link.
    for link in root.findall("link"):
        logger.info("Converting link %s", link.get("name"))
        if (visual := link.find("visual")) is not None:
            if (origin := visual.find("origin")) is not None:
                if (xyz := origin.get("xyz")) is not None:
                    origin.set("xyz", flstrs(xyz))
                if (rpy := origin.get("rpy")) is not None:
                    origin.set("rpy", flstrs(rpy))
            if (material := visual.find("material")) is not None:
                if (color := material.find("color")) is not None:
                    if (rgba := color.get("rgba")) is not None:
                        color.set("rgba", flstrs(rgba))
        if (collision := link.find("collision")) is not None:
            if (origin := collision.find("origin")) is not None:
                if (xyz := origin.get("xyz")) is not None:
                    origin.set("xyz", flstrs(xyz))
                if (rpy := origin.get("rpy")) is not None:
                    origin.set("rpy", flstrs(rpy))
        if (inertial := link.find("inertial")) is not None:
            if (mass := inertial.find("mass")) is not None:
                if (value := mass.get("value")) is not None:
                    mass.set("value", flstr(float(value)))
            if (origin := inertial.find("origin")) is not None:
                if (xyz := origin.get("xyz")) is not None:
                    origin.set("xyz", flstrs(xyz))
                if (rpy := origin.get("rpy")) is not None:
                    origin.set("rpy", flstrs(rpy))
            if (inertia := inertial.find("inertia")) is not None:
                if (ixx := inertia.get("ixx")) is not None:
                    inertia.set("ixx", flstr(float(ixx)))
                if (iyy := inertia.get("iyy")) is not None:
                    inertia.set("iyy", flstr(float(iyy)))
                if (izz := inertia.get("izz")) is not None:
                    inertia.set("izz", flstr(float(izz)))
                if (ixy := inertia.get("ixy")) is not None:
                    inertia.set("ixy", flstr(float(ixy)))
                if (ixz := inertia.get("ixz")) is not None:
                    inertia.set("ixz", flstr(float(ixz)))
                if (iyz := inertia.get("iyz")) is not None:
                    inertia.set("iyz", flstr(float(iyz)))

    save_xml(urdf_path, tree)


def main() -> None:
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument("urdf_path", type=Path)
    args = parser.parse_args()

    convert_floats_to_consistent_types(args.urdf_path)


if __name__ == "__main__":
    main()



"""Defines a utility function for caching requests."""

import datetime
import logging
from pathlib import Path
from typing import Any, Callable, Coroutine, TypeVar

logger = logging.getLogger(__name__)

T = TypeVar("T")


class Cacher:
    def __init__(self, cache_dir: str | Path, invalidate_after_n_minutes: float | None = None) -> None:
        self.cache_map: dict[str, Any] = {}
        self.cache_dir = Path(cache_dir).expanduser().resolve()
        self.cache_dir.mkdir(parents=True, exist_ok=True)
        self.invalidate_after_n_minutes = invalidate_after_n_minutes

    def _run_fn(
        self,
        file_path: Path,
        invalidate_after_n_minutes: float | None = None,
    ) -> bool:
        if not file_path.exists():
            return True
        if file_path.stat().st_size == 0:
            return True
        if invalidate_after_n_minutes is None:
            invalidate_after_n_minutes = self.invalidate_after_n_minutes
        if invalidate_after_n_minutes is None:
            return False
        modified_time = datetime.datetime.fromtimestamp(file_path.stat().st_mtime)
        rerun_fn = (datetime.datetime.now() - modified_time).seconds > invalidate_after_n_minutes * 60
        if rerun_fn:
            logger.warning("Cache invalidated after %.2f minutes", invalidate_after_n_minutes)
            file_path.unlink()
        return rerun_fn

    async def __call__(
        self,
        cache_key: str,
        get_fn: Callable[[], Coroutine[Any, Any, T]],
        from_json_fn: Callable[[str], T],
        to_json_fn: Callable[[T], str],
        invalidate_after_n_minutes: float | None = None,
    ) -> T:
        """Gets an item by calling an API or by retrieving it from the cache.

        Args:
            cache_dir: The directory to store the cache.
            cache_key: The key to use for the cache.
            get_fn: The function to call to get the item.
            from_json_fn: The function to call to parse the item from JSON.
            to_json_fn: The function to call to serialize the item to JSON.
            invalidate_after_n_minutes: The number of minutes after which the cache should be invalidated.

        Returns:
            The item.
        """
        if cache_key in self.cache_map:
            return self.cache_map[cache_key]
        cache_path = self.cache_dir / f"{cache_key}.json"
        if self._run_fn(cache_path, invalidate_after_n_minutes):
            item = await get_fn()
            with open(cache_path, "w") as f:
                f.write(to_json_fn(item))
            self.cache_map[cache_key] = item
        else:
            with open(cache_path) as f:
                item = from_json_fn(f.read())
            self.cache_map[cache_key] = item
        return item



"""Defines a pass to remove internal geometry from meshes."""

import argparse
import logging
from pathlib import Path

from onshape.passes.utils import iter_meshes

logger = logging.getLogger(__name__)


def remove_internal_geometry(mesh_path: Path) -> None:
    """Removes internal geometry from a mesh file by keeping only the outer shell.

    Args:
        mesh_path: Path to the mesh file to process.
    """
    try:
        import open3d as o3d
    except ImportError:
        raise ImportError(
            "Open3D is required to run this script. Install it with `pip install "
            "'onshape[open3d]'` to install the required dependencies."
        )

    mesh = o3d.io.read_triangle_mesh(str(mesh_path))

    # Count initial triangles
    initial_triangles = len(mesh.triangles)

    # Ensure normals are computed
    mesh.compute_vertex_normals()

    # Clean up the resulting mesh
    mesh.remove_degenerate_triangles()
    mesh.remove_duplicated_triangles()
    mesh.remove_duplicated_vertices()
    mesh.remove_non_manifold_edges()

    # Count final triangles and log if any were removed
    final_triangles = len(mesh.triangles)
    if final_triangles < initial_triangles:
        logger.info(
            "Removed %d triangles from %s (%.1f%% reduction)",
            initial_triangles - final_triangles,
            mesh_path.name,
            100 * (initial_triangles - final_triangles) / initial_triangles,
        )

    # Save the mesh back
    match mesh_path.suffix.lower():
        case ".ply":
            o3d.io.write_triangle_mesh(str(mesh_path), mesh, write_ascii=True)
        case ".stl":
            mesh.compute_vertex_normals()
            o3d.io.write_triangle_mesh(str(mesh_path), mesh)
        case _:
            o3d.io.write_triangle_mesh(str(mesh_path), mesh)


def remove_internal_geometries_from_urdf(urdf_path: Path) -> None:
    """Removes internal geometries from all meshes in a URDF.

    Args:
        urdf_path: The path to the URDF file.
    """
    # Iterate through each link in the URDF and process the meshes
    for _, (_, visual_mesh_path), (_, collision_mesh_path) in iter_meshes(urdf_path):
        for mesh_path in list({visual_mesh_path, collision_mesh_path}):
            if mesh_path is not None:
                remove_internal_geometry(mesh_path)


def main() -> None:
    parser = argparse.ArgumentParser(description="Remove internal geometries from meshes in a URDF.")
    parser.add_argument("urdf_path", type=Path, help="The path to the URDF file.")
    args = parser.parse_args()

    remove_internal_geometries_from_urdf(args.urdf_path)


if __name__ == "__main__":
    # python -m onshape.passes.remove_internal_geometries
    main()



"""Defines utility functions used by the project."""

from typing import Any, Coroutine, TypeVar

T = TypeVar("T")


async def catch_error(coro: Coroutine[Any, Any, T]) -> tuple[T | None, Exception | None]:
    try:
        return await coro, None
    except Exception as e:
        return None, e



"""Tests mesh conversion options."""

from pathlib import Path
from typing import get_args

import pytest

from onshape.utils.mesh import MeshType, load_file, save_file, stl_to_fmt


@pytest.mark.parametrize("mesh_type", get_args(MeshType))
def test_stl_to_type(mesh_type: MeshType, tmpdir: Path) -> None:
    if mesh_type == "stl":
        return
    fmt_path = Path(tmpdir / f"random.{mesh_type}")
    stl_to_fmt("tests/data/random.stl", fmt_path)
    assert fmt_path.exists()

    stl_mesh = load_file("tests/data/random.stl")
    fmt_mesh = load_file(fmt_path)

    assert stl_mesh == fmt_mesh

    # Tests saving as the target format.
    fmt_path = Path(tmpdir / f"random.{mesh_type}")
    save_file(stl_mesh, fmt_path)
    assert fmt_path.exists()
    fmt_mesh = load_file(fmt_path)
    assert stl_mesh == fmt_mesh



"""Defines helper functions for MJCF accessors."""

import logging
from dataclasses import dataclass, field
from typing import TYPE_CHECKING, cast, get_args

from onshape.formats.common import ActuatorMetadata, JointMetadata

logger = logging.getLogger(__name__)

if TYPE_CHECKING:
    from urdf2mjcf.convert import ConversionMetadata as ConversionMetadataRef


@dataclass
class CollisionParams:
    condim: int = field(default=6)
    contype: int = field(default=0)
    conaffinity: int = field(default=1)
    solref: list[float] = field(default_factory=lambda: [0.005, 1.0])
    solimp: list[float] = field(default_factory=lambda: [0.99, 0.999, 0.00001])
    friction: list[float] = field(default_factory=lambda: [0.8, 0.02, 0.01])


@dataclass
class ImuSensor:
    body_name: str = field()
    pos: list[float] | None = field(default=None)  # (x, y, z)
    rpy: list[float] | None = field(default=None)  # (roll, pitch, yaw)
    acc_noise: float | None = field(default=None)
    gyro_noise: float | None = field(default=None)
    mag_noise: float | None = field(default=None)


@dataclass
class ForceSensor:
    body_name: str = field()
    site_name: str = field()
    name: str | None = field(default=None)
    noise: float | None = field(default=None)


@dataclass
class ExplicitFloorContacts:
    contact_links: list[str] = field(default_factory=lambda: [])
    class_name: str = field(default="collision")


@dataclass
class CollisionGeometry:
    name: str = field()
    collision_type: str = field()
    sphere_radius: float = field(default=0.01)
    axis_order: tuple[int, int, int] = field(default=(0, 1, 2))
    flip_axis: bool = field(default=False)
    offset_x: float = field(default=0.0)
    offset_y: float = field(default=0.0)
    offset_z: float = field(default=0.0)


@dataclass
class ConversionMetadata:
    suffix: str | None = field(default=None)
    freejoint: bool = field(default=True)
    collision_params: CollisionParams = field(default_factory=lambda: CollisionParams())
    imus: list[ImuSensor] = field(default_factory=lambda: [])
    force_sensors: list[ForceSensor] = field(default_factory=lambda: [])
    collision_geometries: list[CollisionGeometry] = field(default_factory=lambda: [])
    explicit_contacts: ExplicitFloorContacts | None = field(default_factory=ExplicitFloorContacts)
    remove_redundancies: bool = field(default=True)
    floating_base: bool = field(default=True)
    maxhullvert: int = field(default=64)
    angle: str = field(default="radian")
    floor_name: str = field(default="floor")
    add_floor: bool = field(default=False)
    backlash: float | None = field(default=None)
    backlash_damping: float = field(default=0.01)
    height_offset: float = field(default=0.0)
    visualize_collision_meshes: bool = field(default=True)
    joint_name_to_metadata: dict[str, JointMetadata] | None = field(default=None)
    actuator_type_to_metadata: dict[str, ActuatorMetadata] | None = field(default=None)


def convert_to_mjcf_metadata(metadata: ConversionMetadata) -> "ConversionMetadataRef":
    try:
        import urdf2mjcf

        logger.debug("urdf2mjcf version: %s", urdf2mjcf.__version__)

    except ImportError as e:
        raise ImportError(
            "urdf2mjcf is required to run this script. Install it with `pip install "
            "'onshape[mujoco]'` to install the required dependencies."
        ) from e

    from urdf2mjcf.model import (
        ActuatorMetadata as ActuatorMetadataRef,
        Angle,
        CollisionGeometry as CollisionGeometryRef,
        CollisionParams as CollisionParamsRef,
        CollisionType,
        ConversionMetadata as ConversionMetadataRef,
        ExplicitFloorContacts as ExplicitFloorContactsRef,
        ForceSensor as ForceSensorRef,
        ImuSensor as ImuSensorRef,
        JointMetadata as JointMetadataRef,
    )

    if metadata.angle not in get_args(Angle):
        raise ValueError(f"Invalid angle type: {metadata.angle}. Must be one of {get_args(Angle)}")

    for cg in metadata.collision_geometries:
        if not hasattr(CollisionType, cg.collision_type.upper()):
            raise ValueError(f"Bad collision type: {cg.collision_type}. Must be in {CollisionType.__members__.keys()}")

    joint_name_to_metadata = (
        {name: JointMetadataRef.from_dict(vars(param)) for name, param in metadata.joint_name_to_metadata.items()}
        if metadata.joint_name_to_metadata
        else None
    )

    actuator_type_to_metadata = (
        {typ: ActuatorMetadataRef.from_dict(vars(param)) for typ, param in metadata.actuator_type_to_metadata.items()}
        if metadata.actuator_type_to_metadata
        else None
    )

    return ConversionMetadataRef(
        freejoint=metadata.freejoint,
        collision_params=CollisionParamsRef(
            condim=metadata.collision_params.condim,
            contype=metadata.collision_params.contype,
            solref=metadata.collision_params.solref,
            friction=metadata.collision_params.friction,
        ),
        imus=[
            ImuSensorRef(
                body_name=imu.body_name,
                pos=imu.pos,
                rpy=imu.rpy,
                acc_noise=imu.acc_noise,
                gyro_noise=imu.gyro_noise,
                mag_noise=imu.mag_noise,
            )
            for imu in metadata.imus
        ],
        force_sensors=[
            ForceSensorRef(
                body_name=fs.body_name,
                site_name=fs.site_name,
                name=fs.name,
                noise=fs.noise,
            )
            for fs in metadata.force_sensors
        ],
        collision_geometries=[
            CollisionGeometryRef(
                name=cg.name,
                collision_type=getattr(CollisionType, cg.collision_type.upper()),
                sphere_radius=cg.sphere_radius,
                axis_order=cg.axis_order,
                flip_axis=cg.flip_axis,
                offset_x=cg.offset_x,
                offset_y=cg.offset_y,
                offset_z=cg.offset_z,
            )
            for cg in metadata.collision_geometries
        ],
        explicit_contacts=(
            None
            if metadata.explicit_contacts is None
            else ExplicitFloorContactsRef(
                contact_links=metadata.explicit_contacts.contact_links,
                class_name=metadata.explicit_contacts.class_name,
            )
        ),
        remove_redundancies=metadata.remove_redundancies,
        floating_base=metadata.floating_base,
        maxhullvert=metadata.maxhullvert,
        angle=cast(Angle, metadata.angle),
        floor_name=metadata.floor_name,
        add_floor=metadata.add_floor,
        backlash=metadata.backlash,
        backlash_damping=metadata.backlash_damping,
        height_offset=metadata.height_offset,
        visualize_collision_meshes=metadata.visualize_collision_meshes,
        joint_name_to_metadata=joint_name_to_metadata,
        actuator_type_to_metadata=actuator_type_to_metadata,
    )



"""Defines a pass to update the names in a URDF."""

import argparse
import logging
import re
import xml.etree.ElementTree as ET
from pathlib import Path
from typing import Mapping

from onshape.formats.common import save_xml

logger = logging.getLogger(__name__)


def find_not_null(element: ET.Element, key: str) -> ET.Element:
    value = element.find(key)
    if value is None:
        raise ValueError(f"{key} not found in {element.tag}")
    return value


def update_urdf_names(
    urdf_path: Path,
    joint_name_map: Mapping[str, str] | None = None,
    link_name_map: Mapping[str, str] | None = None,
) -> None:
    """Updates the names in a URDF.

    Args:
        urdf_path: The path to the URDF file.
        joint_name_map: A mapping from old joint names to new joint names.
        link_name_map: A mapping from old link names to new link names.
    """
    if joint_name_map is None:
        joint_name_map = {}
    if link_name_map is None:
        link_name_map = {}

    urdf_tree = ET.parse(urdf_path)
    root = urdf_tree.getroot()

    # By default, the link and joint names are quite verbose - they will look
    # something like "link:Upper_Body_(1):Torso_(1):Battery_Container_(1)".
    # Instead, we want to just use "Battery Container". However, we need to
    # make sure that there are no collisions with other names in the URDF, so
    # if there are duplicates then we add a suffix to make them unique.
    # Addtionally, we need to keep track of the original names so that we can
    # update all the links and joints which reference those names to be
    # correct.
    name_mapping: dict[str, str] = {}
    name_set: set[str] = set()

    def cleanup_name(name: str) -> str:
        base_name = name.split(":")[-1]
        base_name = re.sub(r"_\(\d+\)", "", base_name)
        base_name = re.sub(r"-", "_", base_name)
        return base_name

    def get_unique_name(name: str, name_set: set[str]) -> str:
        base_name = cleanup_name(name)
        count = 2
        unique_name = base_name
        while unique_name.lower() in name_set:
            unique_name = f"{base_name}_{count}"
            count += 1
        name_set.add(unique_name.lower())
        return unique_name

    for joint in root.findall(".//joint"):
        if "name" in joint.attrib:
            old_name = joint.attrib["name"]
            new_name = joint_name_map.get(old_name, get_unique_name(old_name, name_set))
            name_mapping[old_name] = new_name
            joint.attrib["name"] = new_name

    for link in root.findall(".//link"):
        if "name" in link.attrib:
            old_name = link.attrib["name"]
            new_name = link_name_map.get(old_name, get_unique_name(old_name, name_set))
            name_mapping[old_name] = new_name
            link.attrib["name"] = new_name

    # Cleans up material names.
    for tag_name in ("material", "geometry", "visual", "inertial", "collision"):
        for tag in root.findall(f".//{tag_name}"):
            if "name" in tag.attrib:
                tag.attrib["name"] = get_unique_name(tag.attrib["name"], name_set)

    # Cleans up mesh STL names (note, this requires renaming the mesh files).
    new_mesh_paths: dict[str, Path] = {}
    mesh_name_set: set[str] = set()
    for mesh in root.findall(".//mesh"):
        if "filename" in mesh.attrib:
            mesh_relpath = mesh.attrib["filename"]
            if mesh_relpath not in new_mesh_paths:
                mesh_path = urdf_path.parent / mesh_relpath
                if not mesh_path.exists():
                    raise FileNotFoundError(mesh_path)
                new_mesh_path = mesh_path.with_name(get_unique_name(mesh_path.stem, mesh_name_set) + mesh_path.suffix)
                new_mesh_paths[mesh_relpath] = new_mesh_path
                mesh_path.rename(new_mesh_path)
            mesh.attrib["filename"] = new_mesh_paths[mesh_relpath].relative_to(urdf_path.parent).as_posix()

    # Now we need to update all the other links and joints which reference
    # these names to be correct.
    for joint in root.findall(".//joint"):
        parent = find_not_null(joint, "parent")
        child = find_not_null(joint, "child")
        parent.attrib["link"] = name_mapping[parent.attrib["link"]]
        child.attrib["link"] = name_mapping[child.attrib["link"]]

    # Registers the namespaces to avoid validation errors.
    ET.register_namespace("", "http://wiki.ros.org/urdf")
    ET.register_namespace("visual", "http://wiki.ros.org/collada_urdf")
    ET.register_namespace("collision", "http://wiki.ros.org/collada_urdf")

    save_xml(urdf_path, root)


def main() -> None:
    parser = argparse.ArgumentParser(description="Merge fixed joints in a URDF.")
    parser.add_argument("urdf_path", type=Path, help="The path to the URDF file.")
    args = parser.parse_args()

    update_urdf_names(args.urdf_path)


if __name__ == "__main__":
    # python -m onshape.passes.simplify_meshes
    main()



"""Tests the OnShape schema against sample data."""

import json

from onshape.onshape.schema.assembly import Assembly
from onshape.onshape.schema.features import Features


def test_assembly_schema() -> None:
    assembly_json_path = "tests/data/dummy_assembly.json"
    with open(assembly_json_path) as f:
        assembly_json = json.load(f)
    assembly = Assembly.model_validate(assembly_json)
    assert assembly is not None


def test_features_schema() -> None:
    features_json_path = "tests/data/dummy_features.json"
    with open(features_json_path) as f:
        features_json = json.load(f)
    features = Features.model_validate(features_json)
    assert features is not None


if __name__ == "__main__":
    # python -m tests.test_onshape_schema
    # test_assembly_schema()
    test_features_schema()



# mypy: disable-error-code="import-untyped, import-not-found"
#!/usr/bin/env python
"""Setup script for the project."""

import re

from setuptools import setup

with open("README.md", "r", encoding="utf-8") as f:
    long_description: str = f.read()


with open("onshape/requirements.txt", "r", encoding="utf-8") as f:
    requirements: list[str] = f.read().splitlines()


with open("onshape/requirements-dev.txt", "r", encoding="utf-8") as f:
    requirements_dev: list[str] = f.read().splitlines()


requirements_mujoco = ["mujoco", "urdf2mjcf", "mujoco-python-viewer"]
requirements_open3d = ["open3d"]
requirements_all = requirements_dev + requirements_mujoco + requirements_open3d


with open("onshape/__init__.py", "r", encoding="utf-8") as fh:
    version_re = re.search(r"^__version__ = \"([^\"]*)\"", fh.read(), re.MULTILINE)
assert version_re is not None, "Could not find version in onshape/__init__.py"
version: str = version_re.group(1)


setup(
    name="onshape",
    version=version,
    description="K-Scale's library for programmatically interacting with OnShape",
    author="Benjamin Bolte",
    url="https://github.com/kscalelabs/onshape",
    license_files=("LICENSE",),
    long_description=long_description,
    long_description_content_type="text/markdown",
    python_requires=">=3.11",
    install_requires=requirements,
    tests_require=requirements_dev,
    extras_require={
        "dev": requirements_dev,
        "mujoco": requirements_mujoco,
        "open3d": requirements_open3d,
        "all": requirements_all,
    },
    package_data={
        "onshape": [
            "py.typed",
            "requirements*.txt",
            "scripts/bullet/*.obj",
            "scripts/bullet/*.urdf",
        ],
    },
)



"""Defines the config class."""

import argparse
from dataclasses import dataclass, field
from pathlib import Path
from typing import Self, Sequence, cast

from omegaconf import MISSING, OmegaConf
from omegaconf.errors import ConfigKeyError

from onshape.formats.common import ActuatorMetadata, JointMetadata
from onshape.formats.mjcf import ConversionMetadata
from onshape.onshape.schema.assembly import Key, MatedEntity, MateType


@dataclass
class MimicRelation:
    parent: Key
    multiplier: float


@dataclass
class Joint:
    parent: Key
    child: Key
    parent_entity: MatedEntity
    child_entity: MatedEntity
    mate_type: MateType
    joint_key: Key
    lhs_is_first: bool


@dataclass
class JointLimits:
    z_min_expression: str | None = field(default=None)
    z_max_expression: str | None = field(default=None)
    axial_z_min_expression: str | None = field(default=None)
    axial_z_max_expression: str | None = field(default=None)


@dataclass
class DownloadConfig:
    document_url: str = field(
        default=MISSING,
        metadata={"help": "The URL of the OnShape document."},
    )
    use_assembly_name: bool = field(
        default=False,
        metadata={"help": "Uses the assembly name as the file name, instead of the default name."},
    )
    output_dir: str = field(
        default=MISSING,
        metadata={"help": "The output directory."},
    )
    default_prismatic_joint_effort: float = field(
        default=80.0,
        metadata={"help": "The default effort limit for prismatic joints."},
    )
    default_prismatic_joint_velocity: float = field(
        default=5.0,
        metadata={"help": "The default velocity limit for prismatic joints."},
    )
    default_revolute_joint_effort: float = field(
        default=80.0,
        metadata={"help": "The default effort limit for revolute joints, in Nm."},
    )
    default_revolute_joint_velocity: float = field(
        default=5.0,
        metadata={"help": "The default velocity limit for revolute joints, in rad/s."},
    )
    default_revolute_joint_limits: tuple[float, float] | None = field(
        default=None,
        metadata={"help": "The default joint limits to use for joints without limits."},
    )
    default_prismatic_joint_limits: tuple[float, float] | None = field(
        default=None,
        metadata={"help": "The default joint limits to use for joints without limits."},
    )
    suffix_to_joint_effort: dict[str, float] = field(
        default_factory=lambda: {},
        metadata={"help": "The suffix to joint effort mapping."},
    )
    suffix_to_joint_velocity: dict[str, float] = field(
        default_factory=lambda: {},
        metadata={"help": "The suffix to joint velocity mapping."},
    )
    invalidate_cache_after_n_minutes: float | None = field(
        default=None,
        metadata={"help": "Invalidates the cache after n minutes."},
    )
    disable_mimics: bool = field(
        default=False,
        metadata={"help": "Disables the mimic joints."},
    )
    override_central_node: str | None = field(
        default=None,
        metadata={"help": "The name of the central node to use."},
    )
    min_facet_width: float | None = field(
        default=0.002,
        metadata={"help": "The minimum facet width to use for simplifying meshes."},
    )
    mesh_dir: str = field(
        default="meshes",
        metadata={"help": "The directory to store the meshes."},
    )
    max_concurrent_requests: int = field(
        default=5,
        metadata={"help": "The maximum number of concurrent requests to make."},
    )
    api_post_wait: float = field(
        default=0.0,
        metadata={"help": "The number of seconds to wait after each API request."},
    )
    default_part_mass: float | None = field(
        default=None,
        metadata={"help": "The default mass to use for parts, if mass is missing."},
    )
    min_inertia_value: float = field(
        default=1e-6,
        metadata={"help": "The minimum value to use for the inertia matrix."},
    )
    debug: bool = field(
        default=False,
        metadata={"help": "Enables debug mode."},
    )

    @classmethod
    def from_cli_args(cls, args: Sequence[str]) -> Self:
        # First, parse the document URL and output directory.
        parser = argparse.ArgumentParser()
        parser.add_argument("document_url", help="The URL of the OnShape document.")
        parser.add_argument("-o", "--output-dir", type=str, default="robot", help="The output directory.")
        parser.add_argument("-f", "--config-path", type=Path, default=None, help="The path to the config file.")
        parser.add_argument("-n", "--no-cache", action="store_true", help="Disables caching.")
        parsed_args, remaining_args = parser.parse_known_args(args)
        document_url: str = parsed_args.document_url
        output_dir: str = parsed_args.output_dir
        config_path: Path | None = parsed_args.config_path

        # Next, parses additional config arguments.
        cfg = cast(Self, OmegaConf.structured(cls))
        cfg.document_url = document_url
        cfg.output_dir = output_dir
        if config_path is not None:
            with config_path.open("r") as f:
                file_cfg = OmegaConf.load(f)
                cfg = cast(Self, OmegaConf.merge(cfg, file_cfg))
        try:
            cli_cfg = OmegaConf.from_cli(remaining_args)
            cfg = cast(Self, OmegaConf.merge(cfg, cli_cfg))
        except ConfigKeyError as e:
            raise ValueError(
                "You tried to supply config override arguments that are not valid for this config. If you meant to "
                "specify the output directory, use `-o <output_dir>` or `--output-dir <output_dir>` instead."
            ) from e
        if cfg.invalidate_cache_after_n_minutes is None and parsed_args.no_cache:
            cfg.invalidate_cache_after_n_minutes = 0.0
        return cfg


@dataclass
class PostprocessConfig:
    update_names: bool = field(
        default=True,
        metadata={"help": "Updates the names in the URDF."},
    )
    joint_name_map: dict[str, str] = field(
        default_factory=lambda: {},
        metadata={"help": "The mapping from old joint names to new joint names."},
    )
    exclude_collision_meshes: list[str] = field(
        default_factory=lambda: [],
        metadata={"help": "The names of the collision meshes to exclude."},
    )
    link_name_map: dict[str, str] = field(
        default_factory=lambda: {},
        metadata={"help": "The mapping from old link names to new link names."},
    )
    urdf_path: str = field(
        default=MISSING,
        metadata={"help": "The path to the downloaded URDF."},
    )
    merge_fixed_joints: bool = field(
        default=True,
        metadata={"help": "Merges fixed joints in the URDF."},
    )
    rotate_joints: dict[str, list[float]] | None = field(
        default=None,
        metadata={"help": "The joints to rotate, and the RPY values to rotate them by."},
    )
    ignore_merging_fixed_joints: list[str] = field(
        default_factory=lambda: [],
        metadata={"help": "The names of the fixed joints to ignore when merging (for example, IMU joints)."},
    )
    simplify_meshes: bool = field(
        default=True,
        metadata={"help": "Simplifies the meshes when converting the URDF."},
    )
    remove_internal_geometries: bool = field(
        default=True,
        metadata={"help": "Remove the internal geometries from meshes as a clean-up step."},
    )
    remove_extra_meshes: bool = field(
        default=True,
        metadata={"help": "Removes any extra mesh files in the URDF directory."},
    )
    joint_separation_distance: float | None = field(
        default=None,
        metadata={"help": "If set, separate adjacent joints by this distance, to prevent collisions."},
    )
    add_base_linkage: bool = field(
        default=True,
        metadata={"help": "Adds a base linkage to the URDF."},
    )
    actuators: list[str] | None = field(
        default=None,
        metadata={"help": "The names of the actuators to use."},
    )
    base_xyz: tuple[float, float, float] = field(
        default=(0.0, 0.0, 0.0),
        metadata={"help": "The XYZ to apply to the base linkage to set the robot's position."},
    )
    base_rpy: tuple[float, float, float] = field(
        default=(0.0, 0.0, 0.0),
        metadata={"help": "The RPY to apply to the base linkage to orient the robot."},
    )
    voxel_size: float = field(
        default=0.00002,
        metadata={"help": "The voxel size to use for simplifying meshes."},
    )
    convex_collision_meshes: bool = field(
        default=False,
        metadata={"help": "Creates separate convex hulls for collision geometries."},
    )
    max_mesh_triangles: int | None = field(
        default=None,
        metadata={"help": "The maximum number of triangles to use for meshes."},
    )
    max_convex_collision_mesh_triangles: int | None = field(
        default=None,
        metadata={"help": "The maximum number of triangles to use for convex collision meshes."},
    )
    separate_collision_meshes: bool = field(
        default=False,
        metadata={"help": "Creates separate collision meshes for each link in the URDF."},
    )
    shrink_collision_meshes: dict[str, float] | None = field(
        default=None,
        metadata={"help": "Shrinks the named collision meshes by this factor."},
    )
    move_collision_meshes: dict[str, list[float]] | None = field(
        default=None,
        metadata={"help": "Moves the named collision meshes by this offset."},
    )
    use_collision_meshes_as_visual_meshes: bool = field(
        default=False,
        metadata={"help": "Uses collision meshes as visual meshes - mainly used for debugging."},
    )
    fix_inertias: bool = field(
        default=True,
        metadata={"help": "Ensures that the inertias are positive definite."},
    )
    convert_floats_to_consistent_types: bool = field(
        default=True,
        metadata={"help": "Converts floating point numbers to consistent types."},
    )
    min_inertia_eigval: float = field(
        default=1e-7,
        metadata={"help": "The minimum value to use for the inertia matrix."},
    )
    flip_joints: list[str] | None = field(
        default=None,
        metadata={"help": "The names of the joints to flip."},
    )
    sort_sections: bool = field(
        default=True,
        metadata={"help": "Sorts the sections in the URDF, putting the joints at the end."},
    )
    add_mjcf: bool = field(
        default=True,
        metadata={"help": "Adds the MJCF XML to the package."},
    )
    mjcf_metadata: list[ConversionMetadata] | None = field(
        default=None,
        metadata={"help": "The MJCF metadata(s) to use for the URDF."},
    )
    joint_metadata: dict[str, JointMetadata] | None = field(
        default=None,
        metadata={"help": "The joint metadata to use for the URDF."},
    )
    actuators_metadata: dict[str, ActuatorMetadata] | None = field(
        default=None,
        metadata={"help": "Dictionary of actuator types to actuator parameters."},
    )
    package_tgz: bool = field(
        default=True,
        metadata={"help": "Packages the URDF into a .tgz file."},
    )
    debug: bool = field(
        default=False,
        metadata={"help": "Enables debug mode."},
    )

    @classmethod
    def from_cli_args(cls, args: Sequence[str]) -> Self:
        # First, parse the URDF path.
        parser = argparse.ArgumentParser()
        parser.add_argument("urdf_path", help="The path to the downloaded URDF.")
        parser.add_argument("-f", "--config-path", type=Path, default=None, help="The path to the config file.")
        parsed_args, remaining_args = parser.parse_known_args(args)
        urdf_path: str = parsed_args.urdf_path
        config_path: Path | None = parsed_args.config_path

        # Next, parses additional config arguments.
        cfg = cast(Self, OmegaConf.structured(cls))
        cfg.urdf_path = urdf_path
        if config_path is not None:
            with config_path.open("r") as f:
                file_cfg = OmegaConf.load(f)
                cfg = cast(Self, OmegaConf.merge(cfg, file_cfg))
        try:
            cli_cfg = OmegaConf.from_cli(remaining_args)
            cfg = cast(Self, OmegaConf.merge(cfg, cli_cfg))
        except ConfigKeyError as e:
            raise ValueError(
                "You tried to supply config override arguments that are not valid for this config. If you meant to "
                "specify the output directory, use `-o <output_dir>` or `--output-dir <output_dir>` instead."
            ) from e
        return cfg


@dataclass
class ConverterConfig(DownloadConfig, PostprocessConfig):
    @classmethod
    def from_cli_args(cls, args: Sequence[str]) -> Self:
        # First, parse the document URL and output directory.
        parser = argparse.ArgumentParser()
        parser.add_argument("document_url", help="The URL of the OnShape document.")
        parser.add_argument("-o", "--output-dir", type=str, default="robot", help="The output directory.")
        parser.add_argument("-f", "--config-path", type=Path, default=None, help="The path to the config file.")
        parser.add_argument("-n", "--no-cache", action="store_true", help="Disables caching.")
        parsed_args, remaining_args = parser.parse_known_args(args)
        document_url: str = parsed_args.document_url
        output_dir: str = parsed_args.output_dir
        config_path: Path | None = parsed_args.config_path

        # Next, parses additional config arguments.
        cfg = cast(Self, OmegaConf.structured(cls))
        cfg.document_url = document_url
        cfg.output_dir = output_dir
        if config_path is not None:
            with config_path.open("r") as f:
                file_cfg = OmegaConf.load(f)
                cfg = cast(Self, OmegaConf.merge(cfg, file_cfg))
        try:
            cli_cfg = OmegaConf.from_cli(remaining_args)
            cfg = cast(Self, OmegaConf.merge(cfg, cli_cfg))
        except ConfigKeyError as e:
            raise ValueError(
                "You tried to supply config override arguments that are not valid for this config. If you meant to "
                "specify the output directory, use `-o <output_dir>` or `--output-dir <output_dir>` instead."
            ) from e
        if cfg.invalidate_cache_after_n_minutes is None and parsed_args.no_cache:
            cfg.invalidate_cache_after_n_minutes = 0.0
        return cfg



"""Defines utility functions for working with mesh files."""

from dataclasses import dataclass
from pathlib import Path
from typing import Any, Literal, Union, cast, get_args

import numpy as np
import trimesh
from numpy.typing import NDArray

MeshType = Literal["stl", "obj"]

COLLISION_SUFFIX = ".collision"


@dataclass
class Mesh:
    """Defines a common representation for a mesh.

    Attributes:
        points: The points of the mesh, where each point is a 3D coordinate.
        faces: The faces of the mesh, where each face is a triangle defined
            by the indices of its vertices in the points array.
    """

    points: NDArray
    faces: NDArray

    def __eq__(self, other: Any) -> bool:  # noqa: ANN401
        if not isinstance(other, Mesh):
            return False
        return np.allclose(self.points, other.points) and np.array_equal(self.faces, other.faces)

    def save(self, file_path: Union[str, Path]) -> None:
        save_file(self, file_path)

    @classmethod
    def from_trimesh(cls, mesh: trimesh.Trimesh) -> "Mesh":
        return cls(mesh.vertices, mesh.faces)

    def to_trimesh(self) -> trimesh.Trimesh:
        return trimesh.Trimesh(self.points, self.faces)


def get_mesh_type(file_path: Union[str, Path]) -> MeshType:
    if (file_path_str := str(file_path).lower()) in get_args(MeshType):
        return cast(MeshType, file_path_str)
    ext = Path(file_path).suffix[1:]
    if ext not in get_args(MeshType):
        raise ValueError(f"Unsupported mesh format: '{ext}'")
    return cast(MeshType, ext)


def stl_to_fmt(stl_path: Union[str, Path], output_path: Union[str, Path]) -> None:
    if get_mesh_type(output_path) != "stl":
        save_file(load_file(stl_path), output_path)


def fmt_to_stl(input_path: Union[str, Path], stl_path: Union[str, Path]) -> None:
    if get_mesh_type(input_path) != "stl":
        save_file(load_file(input_path), stl_path)


def load_file(file_path: Union[str, Path]) -> Mesh:
    mesh = trimesh.load(file_path)
    if not isinstance(mesh, trimesh.Trimesh):
        raise ValueError(f"Unsupported mesh type: {type(mesh)}")
    return Mesh.from_trimesh(mesh)


def save_file(mesh: Mesh, file_path: Union[str, Path]) -> None:
    mesh.to_trimesh().export(file_path)



"""Sorts the sections in the URDF, putting the links at the end."""

import argparse
import logging
import xml.etree.ElementTree as ET
from pathlib import Path

from onshape.formats.common import save_xml

logger = logging.getLogger(__name__)


def sort_sections(urdf_path: Path) -> None:
    urdf_tree = ET.parse(urdf_path)
    root = urdf_tree.getroot()
    links = root.findall("link")
    for link in links:
        root.remove(link)
        root.append(link)
    save_xml(urdf_path, urdf_tree)


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("urdf_path", type=Path, help="The path to the URDF file.")
    args = parser.parse_args()
    sort_sections(args.urdf_path)
